import logging

import numpy as np
import xarray as xr
import pandas as pd

logger = logging.getLogger("ufs2arco")

def get_mappings(time="time") -> dict:
    return {
        "latitude": _latitude,
        "cos_latitude": _cos_latitude,
        "sin_latitude": _sin_latitude,
        "longitude": _longitude,
        "cos_longitude": _cos_longitude,
        "sin_longitude": _sin_longitude,
        "julian_day": lambda xds: _julian_day(xds, time=time),
        "cos_julian_day": lambda xds: _cos_julian_day(xds, time=time),
        "sin_julian_day": lambda xds: _sin_julian_day(xds, time=time),
        "cos_local_time": lambda xds: _cos_local_time(xds, time=time),
        "sin_local_time": lambda xds: _sin_local_time(xds, time=time),
        "cos_solar_zenith_angle": lambda xds: _cos_solar_zenith_angle(xds, time=time),
        "insolation": lambda xds: _cos_solar_zenith_angle(xds, time=time),
    }

def _latitude(xds: xr.Dataset):
    lat = xds["latitude"]
    lat.attrs["computed_forcing"] = True
    lat.attrs["constant_in_time"] = True
    return lat

def _cos_latitude(xds: xr.Dataset):
    clat = np.cos(xds["latitude"]/180*np.pi)
    clat.attrs["computed_forcing"] = True
    clat.attrs["constant_in_time"] = True
    return clat

def _sin_latitude(xds: xr.Dataset):
    slat = np.sin(xds["latitude"]/180*np.pi)
    slat.attrs["computed_forcing"] = True
    slat.attrs["constant_in_time"] = True
    return slat

def _longitude(xds: xr.Dataset):
    lon = xds["longitude"]
    lon.attrs["computed_forcing"] = True
    lon.attrs["constant_in_time"] = True
    return lon

def _cos_longitude(xds: xr.Dataset):
    clon = np.cos(xds["longitude"]/180*np.pi)
    clon.attrs["computed_forcing"] = True
    clon.attrs["constant_in_time"] = True
    return clon

def _sin_longitude(xds: xr.Dataset):
    slon = np.sin(xds["longitude"]/180*np.pi)
    slon.attrs["computed_forcing"] = True
    slon.attrs["constant_in_time"] = True
    return slon

def _julian_day(xds: xr.Dataset, time="time"):
    ptime = pd.to_datetime(xds[time])
    pyears = pd.to_datetime(xds[time].dt.year.astype(str))
    delta = ptime - pyears
    jday = delta.days + delta.seconds / 86400
    jday = xr.DataArray(
        jday,
        coords=xds[time].coords,
        attrs={
            "description": "julian day relative to start of year",
            "computed_forcing": True,
            "constant_in_time": False,
        },
    )
    return jday

def _cos_julian_day(xds: xr.Dataset, time="time"):
    jd = _julian_day(xds, time=time)
    cjd = np.cos(jd/365.25*2*np.pi)
    cjd.attrs["description"] = f"cosine of {jd.attrs['description']}"
    cjd.attrs["computed_forcing"] = True
    cjd.attrs["constant_in_time"] = False
    return cjd

def _sin_julian_day(xds: xr.Dataset, time="time"):
    jd = _julian_day(xds, time=time)
    sjd = np.sin(jd/365.25*2*np.pi)
    sjd.attrs["description"] = f"sine of {jd.attrs['description']}"
    sjd.attrs["computed_forcing"] = True
    sjd.attrs["constant_in_time"] = False
    return sjd

def _local_time(xds: xr.Dataset, time="time"):
    pd_time = pd.to_datetime(xds[time].values)
    delta = pd_time - pd_time.normalize() # gets the delta in hours
    hours = delta.seconds / 86400 * 24 # now this is e.g. 12 at 12z
    hours = xr.DataArray(hours, coords=xds[time].coords)
    local_time = (xds["longitude"] / 360 * 24 + hours) % 24
    local_time.attrs["description"] = "relative local time in hours, from time in UTC & longitude"
    local_time.attrs["computed_forcing"] = True
    local_time.attrs["constant_in_time"] = False
    return local_time

def _cos_local_time(xds: xr.Dataset, time="time"):
    lt = _local_time(xds, time=time)
    clt = np.cos(lt/24*2*np.pi)
    clt.attrs["description"] = f"cosine of {lt.attrs['description']}"
    clt.attrs["computed_forcing"] = True
    clt.attrs["constant_in_time"] = False
    return clt

def _sin_local_time(xds: xr.Dataset, time="time"):
    lt = _local_time(xds, time=time)
    slt = np.sin(lt/24*2*np.pi)
    slt.attrs["description"] = f"sin of {lt.attrs['description']}"
    slt.attrs["computed_forcing"] = True
    slt.attrs["constant_in_time"] = False
    return slt

def _solar_declination_angle(xds: xr.Dataset, time="time"):
    """Note: this was copied from github.com/ecmwf/earthkit-meteo
    See: src/earthkit/meteo/solar/array/solar.py "solar_declination_angle"

    Returns:
        declination (xr.DataArray): in radians, not in degrees as in earthkit-meteo
        time_correction (xr.DataArray): function of dataset time array
    """
    angle = _julian_day(xds, time=time) / 365.25 * np.pi * 2
    declination = np.deg2rad(
        0.396372
        - 22.91327 * np.cos(angle)
        + 4.025430 * np.sin(angle)
        - 0.387205 * np.cos(2 * angle)
        + 0.051967 * np.sin(2 * angle)
        - 0.154527 * np.cos(3 * angle)
        + 0.084798 * np.sin(3 * angle)
    )
    # time correction in [ h.degrees ]
    time_correction = (
        0.004297
        + 0.107029 * np.cos(angle)
        - 1.837877 * np.sin(angle)
        - 0.837378 * np.cos(2 * angle)
        - 2.340475 * np.sin(2 * angle)
    )
    return declination, time_correction

def _cos_solar_zenith_angle(xds: xr.Dataset, time="time"):
    """Note: this was copied from github.com/ecmwf/earthkit-meteo
    See: src/earthkit/meteo/solar/array/solar.py "cos_solar_zenith_angle"
    """

    declination, time_correction = _solar_declination_angle(xds, time=time)
    rlat = np.deg2rad(xds["latitude"])

    sindec_sinlat = np.sin(declination) * np.sin(rlat)
    cosdec_coslat = np.cos(declination) * np.cos(rlat)

    # solar hour angle [h.deg]
    solar_angle = np.deg2rad(
        (xds[time].dt.hour - 12) * 15 + xds["longitude"] + time_correction
    )
    zenith_angle = sindec_sinlat + cosdec_coslat * np.cos(solar_angle)
    zenith_angle = zenith_angle.where(zenith_angle > 0., 0.)
    zenith_angle.attrs["description"] = "cosine of solar zenith angle"
    zenith_angle.attrs["attribution"] = "this is computed exactly as in github.com/ecmwf/earthkit-meteo (see src/earthkit/meteo/solar/array/solar.py)"
    zenith_angle.attrs["computed_forcing"] = True
    zenith_angle.attrs["constant_in_time"] = False
    return zenith_angle



import itertools

def batched(iterable, n):
    """A homegrown version of itertools.batched, taken
    directly from `itertools docs <https://docs.python.org/3/library/itertools.html#itertools.batched>`_
    for earlier versions of itertools

    Args:
        iterable (Iterable): loop over this
        n (int): batch size

    Returns:
        batch (Iterable): a batch of size n from the iterable
    """
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch



try:
    from mpi4py import MPI
except:
    pass

import os
import logging
import yaml
from datetime import datetime

import numpy as np
import xarray as xr
import pandas as pd
import zarr

from ufs2arco.log import setup_simple_log
from ufs2arco.mpi import MPITopology, SerialTopology
from ufs2arco import sources
from ufs2arco.transforms import Transformer
from ufs2arco import targets
from ufs2arco.datamover import DataMover, MPIDataMover

logger = logging.getLogger("ufs2arco")

class Driver:
    """A class to manage data movement.

    Attributes:
        config (dict): Configuration dictionary loaded from the YAML file.
        Source, Transformer, Target
        Mover (Type[DataMover] | Type[MPIDataMover]): The data mover class (DataMover or MPIDataMover).

    Methods:
        __init__(config_filename: str): Initializes the Driver object with configuration from the specified YAML file.
        use_mpi: Returns whether MPI should be used based on the configuration.
        source_kwargs: Returns the arguments for initializing the source dataset.
        target_kwargs: Returns the arguments for initializing the target dataset.
        mover_kwargs: Returns the arguments for initializing the mover.
        run(overwrite: bool = False): Runs the data movement process, managing the source, target transformations, and mover.
    """

    required_sections = (
        "mover",
        "directories",
        "source",
        "target",
    )

    recognized_sections = (
        "mover",
        "directories",
        "source",
        "transforms",
        "target",
        "attrs",
    )



    def __init__(self, config_filename: str):
        """Initializes the Driver object with configuration from the specified YAML file.

        Args:
            config_filename (str): Path to the YAML configuration file.

        Raises:
            AssertionError: If required sections or keys are missing in the configuration.
            NotImplementedError: If a source, target, or mover is not recognized.
        """
        with open(config_filename, "r") as f:
            self.config = yaml.safe_load(f)

        for key in self.required_sections:
            assert key in self.config, \
                f"Driver.__init__: could not find '{key}' section in yaml"

        unrecognized = []
        for key in self.config.keys():
            if key not in self.recognized_sections:
                unrecognized.append(key)
        if len(unrecognized) > 0:
            raise KeyError(
                f"Driver.__init__: Unrecognized config sections: {unrecognized}. The following are recognized: {self.recognized_sections}")

        # the source dataset
        name = self.config["source"]["name"].lower()
        if name not in sources._recognized:
            raise NotImplementedError(f"Driver.__init__: unrecognized data source {name}. Must be one of {sources._recognized}")
        self.SourceDataset = getattr(sources, sources._recognized[name])

        # the target
        name = self.config["target"].get("name", "base")
        name = name.lower()
        if name in ("forecast", "analysis", "base"):
            self.TargetDataset = targets.Target
        elif name == "anemoi":
            self.TargetDataset = targets.Anemoi
        else:
            raise NotImplementedError(f"Driver.__init__: only 'base' and 'anemoi' are implemented")

        for key in ["chunks"]:
            assert key in self.config["target"], \
                f"Driver.__init__: could not find '{key}' in 'target' section of yaml"

        # the mover
        name = self.config["mover"]["name"].lower()
        if name == "datamover":
            self.Mover = DataMover
        elif name == "mpidatamover":
            self.Mover = MPIDataMover
        else:
            raise NotImplementedError(f"Driver.__init__: don't recognize mover = {name}")

        for key in ["cache_dir"]:
            assert key not in self.config["mover"], \
                f"Driver.__init__: '{key}' not allowed in 'mover' section of yaml"

        # directories
        dirs = self.config["directories"]
        for key in ["zarr", "cache"]:
            assert key in dirs, \
                f"Driver.__init__: could not find '{key}' in 'directories' section in yaml"

    @property
    def use_mpi(self) -> bool:
        """Determines if MPI is to be used based on the mover configuration.

        Returns:
            bool: True if MPI is to be used, False otherwise.
        """
        return self.config["mover"]["name"].lower() == "mpidatamover"

    @property
    def source_kwargs(self) -> dict:
        """Returns the arguments for initializing the source dataset.

        Returns:
            dict: The source dataset initialization arguments.
        """
        kw = {key: val for key, val in self.config["source"].items() if key != "name"}
        return kw

    @property
    def target_kwargs(self) -> dict:
        """Returns the arguments for initializing the target dataset.

        Returns:
            dict: The target dataset initialization arguments.
        """
        kw = {key: val for key, val in self.config["target"].items() if key != "name"}
        kw["store_path"] = os.path.expandvars(self.config["directories"]["zarr"])
        return kw

    @property
    def mover_kwargs(self) -> dict:
        """Returns the arguments for initializing the data mover.

        Returns:
            dict: The mover initialization arguments.
        """
        kw = {key: val for key, val in self.config["mover"].items() if key != "name"}

        # optional
        if "start" not in kw.keys():
            kw["start"] = 0

        # enforced options
        kw["cache_dir"] = os.path.expandvars(self.config["directories"]["cache"])
        return kw

    def setup(self, runtype):
        # MPI requires some extra setup
        mover_kwargs = self.mover_kwargs.copy()

        log_dir = os.path.expandvars(self.config["directories"]["logs"])

        if runtype != "create":
            if log_dir.endswith("/"):
                log_dir = log_dir[:-1]
            log_dir += f"-{runtype}"

        if self.use_mpi:
            topo = MPITopology(log_dir=log_dir)
            mover_kwargs["mpi_topo"] = topo

        else:
            topo = SerialTopology(log_dir=log_dir)

        source = self.SourceDataset(**self.source_kwargs)
        # create the transformer
        if "transforms" in self.config.keys():
            self.transformer = Transformer(options=self.config["transforms"])
        else:
            self.transformer = None
        target = self.TargetDataset(source=source, **self.target_kwargs)
        mover = self.Mover(source=source, target=target, transformer=self.transformer, **mover_kwargs)
        return topo, mover, source, target

    def run(self, overwrite: bool = False):
        """Runs the data movement process, managing the datasets and mover.

        This method sets up the datasets, creates the container, and loops through
        batches to move data to the specified store path (Zarr format).

        Args:
            overwrite (bool, optional): Whether to overwrite the existing container.
                Defaults to False.
        """
        topo, mover, source, target = self.setup(runtype="create")

        # create container, only if mover start is not 0
        if mover.start == 0:
            container_kwargs = {"mode": "w"} if overwrite else {}
            mover.create_container(**container_kwargs)

        # loop through batches
        n_batches = len(mover)
        missing_dims = []
        for batch_idx in range(mover.start, n_batches):

            xds = next(mover)


            # xds is None if MPI rank looks for non existent indices (i.e., last batch scenario)
            # len(xds) == 0 if we couldn't find the file we were looking for
            has_content = xds is not None and len(xds) > 0
            if has_content:

                xds = xds.reset_coords(drop=True)
                region = mover.find_my_region(xds)
                xds.to_zarr(target.store_path, region=region)
                mover.clear_cache(batch_idx)

            elif xds is not None:

                # we couldn't find the file, keep track of it
                batch_indices = mover.get_batch_indices(batch_idx)
                for these_dims in batch_indices:
                    missing_dims.append(these_dims)

            logger.info(f"Done with batch {batch_idx+1} / {n_batches}")

        topo.barrier()
        logger.info(f"Done moving the data\n")

        self.report_missing_data(topo, source, target, missing_dims)
        target.finalize(topo)
        self.finalize_attributes(topo, target)

    def patch(self):

        topo, mover, source, target = self.setup(runtype="patch")
        missing_dims = _open_patch_yaml(self.get_missing_data_path(target.store_path))

        logger.info(f"Starting patch workflow with missing_dims\n{missing_dims}\n")

        # hacky for now, reset the mover's sample_indices to be the missing dims
        mover.sample_indices = missing_dims
        mover.restart(idx=0)

        missing_again = list()
        n_batches = len(mover)
        for batch_idx, xds in enumerate(mover):

            # xds is None if MPI rank looks for non existent indices (i.e., last batch scenario)
            # len(xds) == 0 if we couldn't find the file we were looking for
            has_content = xds is not None and len(xds) > 0
            if has_content:

                xds = xds.reset_coords(drop=True)
                region = mover.find_my_region(xds)
                xds.to_zarr(target.store_path, region=region)
                mover.clear_cache(batch_idx)

            elif xds is not None:

                batch_indices = mover.get_batch_indices(batch_idx)
                for these_dims in batch_indices:
                    missing_again.append(these_dims)

            logger.info(f"Done with batch {batch_idx+1} / {n_batches}")

        topo.barrier()
        logger.info(f"Done moving the data\n")

        self.report_missing_data(topo, source, target, missing_again)
        target.finalize(topo)
        self.finalize_attributes(topo, target)


    def finalize_attributes(self, topo, target):

        logger.info(f"Storing the recipe and anything from the 'attrs' section in zarr store attributes")
        if topo.is_root:
            zds = zarr.open(target.store_path, mode="a")
            zds.attrs["recipe"] = self.config
            if "attrs" in self.config.keys():
                for key, val in self.config["attrs"].items():
                    zds.attrs[key] = val

            zds.attrs["latest_write_timestamp"] = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            # just in case
            zarr.consolidate_metadata(target.store_path)
        topo.barrier()
        logger.info(f"🚀🚀🚀 Dataset is ready for launch at: {target.store_path}")


    def get_missing_data_path(self, store_path) -> str:
        directory, zstore = os.path.split(store_path)
        return f"{directory}/missing.{zstore}.yaml"

    def report_missing_data(self, topo, source, target, missing_dims):

        missing_dims = topo.gather(missing_dims)

        if topo.is_root:
            # collect the list of lists into one list of dicts
            missing_dims = [item for sublist in missing_dims for item in sublist]

            if len(missing_dims) > 0:
                # sort it by t0 or time
                # then convert numpy or pandas types to int/str etc
                missing_dims = sorted(missing_dims, key=_get_time)
                missing_dims = [_convert_types_to_yaml(d) for d in missing_dims]

                target.handle_missing_data(missing_dims)

                missing_data_yaml = self.get_missing_data_path(target.store_path)
                with open(missing_data_yaml, "w") as f:
                    yaml.dump(missing_dims, stream=f)

                logger.warning(f"⚠️  Some data are missing.")
                logger.warning(f"⚠️  The missing dimension combos, i.e., {source.sample_dims}")
                logger.warning(f"⚠️  were written to: {missing_data_yaml}")
                logger.warning(f"If you know the files are actually available, You can run")
                logger.warning(f"\tpython -c 'import ufs2arco; ufs2arco.Driver(\"/path/to/your/original/recipe.yaml\").patch()'")
                logger.warning(f"to try getting the data again\n")

# some utilities for handling missing data
def _get_time(d):
    return d.get("t0", d.get("time", None))

def _convert_types_to_yaml(d):
    d = d.copy()
    # Convert pd.Timestamp to string
    if "t0" in d and isinstance(d["t0"], pd.Timestamp):
        d["t0"] = str(d["t0"])
    if "time" in d and isinstance(d["time"], pd.Timestamp):
        d["time"] = str(d["time"])
    # Convert numpy integers to Python int
    for key in ["fhr", "member"]:
        if key in d and isinstance(d[key], np.integer):
            d[key] = int(d[key])
    return d

def _open_patch_yaml(yamlpath):

    with open(yamlpath, "r") as f:
        missing_dims = yaml.safe_load(f)
    # Convert string -> pd.Timestamp
    for i, d in enumerate(missing_dims):
        if "t0" in d and isinstance(d["t0"], str):
            missing_dims[i]["t0"] = pd.Timestamp(d["t0"])
        if "time" in d and isinstance(d["time"], str):
            missing_dims[i]["time"] = pd.Timestamp(d["time"])
    return missing_dims



"""This can be used to recreate the available variables tables if more variables are added"""
import yaml

for forecast_system in ["gefs", "gfs", "hrrr"]:

    with open(f"../ufs2arco/sources/reference.{forecast_system}.yaml") as f:
        data = yaml.safe_load(f)

    with open(f"variables.{forecast_system}.rst", "w") as out:
        # Write the table header
        out.write(f".. list-table:: Variables from {forecast_system.upper()} available in ufs2arco\n")
        if forecast_system == "gfs":
            out.write("   :widths: 18 50 12\n")
        else:
            out.write("   :widths: 18 50\n")
        out.write("   :header-rows: 1\n\n")
        out.write("   * - Variable\n")
        out.write("     - Long Name\n")

        if forecast_system == "gfs":
            out.write("     - Forecast Only\n")
        for var, attrs in data.items():
            long_name = attrs.get("long_name", "")
            out.write(f"   * - ``{var}``\n")
            out.write(f"     - {long_name}\n")

            if forecast_system == "gfs":
                forecast_only = str(attrs.get("forecast_only", None))
                out.write(f"     - {forecast_only}\n")



"""A class to pick up the pieces from a failed ReplayMover job

1. figure out the gaps in the dataset by finding the cycle dates that did not finish
2. create a list with all of these cycle dates
3. method to convert these cycle dates to time
4. run through the cycle list, one at a time.

"""

from os.path import isdir
from shutil import rmtree
import logging
import numpy as np
import xarray as xr
import pandas as pd

from ufs2arco import FV3Dataset, Timer

from replay_mover import ReplayMover1Degree, ReplayMoverQuarterDegree


class ReplayFiller:
    """Fill in missing data due to failed jobs from ReplayMover

    Args:
        mover (ReplayMover): the ReplayMover object used. Note that it uses this object to determine how many jobs (nodes) to use and where to store cache, etc.
    """
    # from slurm and the data
    cycles_to_fill = np.concatenate([
        pd.date_range("1995-09-26T00", "1995-12-26T18", freq="6h"),
        pd.date_range("1997-01-07T00", "1997-12-21T00", freq="6h"),
        pd.date_range("1999-09-23T06", "1999-12-16T06", freq="6h"),
        pd.date_range("2000-12-17T12", "2001-12-10T06", freq="6h"),
        pd.date_range("2003-06-29T12", "2003-12-05T12", freq="6h"),
        pd.date_range("2005-07-27T18", "2005-11-29T18", freq="6h"),
        pd.date_range("2007-05-11T00", "2007-11-24T18", freq="6h"),
        pd.date_range("2009-09-12T00", "2009-11-19T00", freq="6h"),
        pd.date_range("2011-10-30T06", "2011-11-14T06", freq="6h"),
        pd.date_range("2015-03-14T12", "2015-11-03T12", freq="6h"),
        pd.date_range("2021-06-21T00", "2021-10-18T00", freq="6h"),
        pd.date_range("2022-10-12T06", "2023-10-13T06", freq="6h"),
    ])

    # this was a necessary quick 2nd fill
    #cycles_to_fill = np.array([
    #    np.datetime64("1997-11-23T00"),
    #    np.datetime64("2015-03-15T06"),
    #    np.datetime64("2015-04-01T06"),
    #])


    @property
    def xcycles_to_fill(self):
        """cycles_to_fill in xarray.DataArray form for time accessors"""
        return xr.DataArray(self.cycles_to_fill, coords={"cycles": self.cycles_to_fill}, dims="cycles")

    @property
    def splits(self):
        """Figure out the subset of cycles each job will be working on"""
        return [int(x) for x in np.linspace(0, len(self.xcycles_to_fill), self.n_jobs+1)]

    @property
    def n_jobs(self):
        return min(self.mover.n_jobs, len(self.cycles_to_fill))

    def __init__(self, mover):
        assert isinstance(mover, (ReplayMover1Degree, ReplayMoverQuarterDegree))
        self.mover = mover


    def my_cycles_to_fill(self, job_id):
        """The cycle timestamps for this job

        Args:
            job_id (int): the slurm job id, determines cache storage location

        Returns:
            cycles_datetime (List[datetime]): with the cycle numbers to be processed by this slurm job
        """
        slices = [slice(st, ed) for st, ed in zip(self.splits[:-1], self.splits[1:])]
        xda = self.xcycles_to_fill.isel(cycles=slices[job_id])
        cycles_datetime = self.mover.npdate2datetime(xda)
        return cycles_datetime


    def run(self, job_id):

        walltime = Timer()
        localtime = Timer()

        walltime.start(f"Starting Job {job_id}")
        replay = FV3Dataset(path_in=self.mover.cached_path, config_filename=self.mover.config_filename)

        for cycle in self.my_cycles_to_fill(job_id):

            localtime.start(f"Reading {str(cycle)}")
            try:
                xds = replay.open_dataset(cycle, **self.mover.ods_kwargs(job_id))

                index = list(self.mover.xtime.values).index(xds.time.values[0])
                tslice = slice(index, index+2)
                print("index = ", index)

                replay.store_dataset(
                    xds,
                    region={
                        "time": tslice,
                        "pfull": slice(None, None),
                        "grid_yt": slice(None, None),
                        "grid_xt": slice(None, None),
                    },
                    storage_options=self.mover.storage_options,
                )

                # This is a hacky way to clear the cache, since we don't create a filesystem object
                del xds
                if isdir(self.mover.cache_storage(job_id)):
                    rmtree(self.mover.cache_storage(job_id), ignore_errors=True)

            except Exception as e:
                logging.exception(e)
                print(f"ReplayFiller.run({job_id}): Failed to store {str(cycle)}")
                pass
            localtime.stop()

        walltime.stop("Total Walltime")



"""Compute geopotential from Replay dataset,
append it back to the original or store it locally depending on inputs

Note: this was heavily borrowed from this xarray-beam example:
https://github.com/google/xarray-beam/blob/main/examples/era5_climatology.py
"""

from typing import Tuple

from absl import app
from absl import flags
import logging
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.runners.dask.dask_runner import DaskRunner
import numpy as np
import xarray as xr
import xarray_beam as xbeam

from ufs2arco import Layers2Pressure
from verify_geopotential import setup_log
from localzarr import ChunksToZarr


INPUT_PATH = flags.DEFINE_string('input_path', None, help='Input Zarr path')
OUTPUT_PATH = flags.DEFINE_string('output_path', None, help='Output Zarr path')
RUNNER = flags.DEFINE_string('runner', "DirectRunner", 'beam.runners.Runner')
TIME_LENGTH = flags.DEFINE_integer('time_length', None, help="Number of time slices to use for debugging")
NUM_WORKERS = flags.DEFINE_integer('num_workers', None, help="Number of workers for the runner")
NUM_THREADS = flags.DEFINE_integer('num_threads', None, help="Passed to ChunksToZarr")

def calc_geopotential(
    key: xbeam.Key,
    xds: xr.Dataset,
) -> Tuple[xbeam.Key, xr.Dataset]:
    """Return dataset with geopotential field, that's it"""

    lp = Layers2Pressure()
    xds = xds.rename({"pfull": "level"})
    prsl = lp.calc_layer_mean_pressure(xds["pressfc"], xds["tmp"], xds["spfh"], xds["delz"])

    newds = xr.Dataset()
    newds["geopotential"] = lp.calc_geopotential(xds["hgtsfc"], xds["delz"])
    newds = newds.rename({"level": "pfull"})
    return key, newds

def main(argv):

    setup_log()
    path = INPUT_PATH.value
    kwargs = {}

    if "gs://" in path or "gcs://" in path:
        kwargs["storage_options"] = {"token": "anon"}

    source_dataset, source_chunks = xbeam.open_zarr(path, **kwargs)
    source_dataset = source_dataset.drop_vars(["cftime", "ftime"])
    if TIME_LENGTH.value is not None:
        source_dataset = source_dataset.isel(time=slice(int(TIME_LENGTH.value)))

    # create template
    tds = source_dataset[["tmp"]].rename({"tmp": "geopotential"})
    tds["geopotential"].attrs = {
        "units": "m**2 / s**2",
        "description": "Diagnosed using ufs2arco.Layers2Pressure.calc_geopotential",
        "long_name": "geopotential height",
    }
    input_chunks = {k: source_chunks[k] if k != "pfull" else 127 for k in tds["geopotential"].dims}
    output_chunks = {k: source_chunks[k] for k in tds["geopotential"].dims}

    template = xbeam.make_template(tds)
    storage_options = None
    if "gs://" in OUTPUT_PATH.value:
        storage_options = {"token": "/contrib/Tim.Smith/.gcs/replay-service-account.json"}

    pipeline_kwargs = {}
    if NUM_WORKERS.value is not None:
        pipeline_kwargs["options"]=PipelineOptions(
            direct_num_workers=NUM_WORKERS.value,
        )

    with beam.Pipeline(runner=RUNNER.value, argv=argv, **pipeline_kwargs) as root:
        (
            root
            | xbeam.DatasetToChunks(source_dataset, input_chunks, num_threads=NUM_THREADS.value)
            | beam.MapTuple(calc_geopotential)
            | ChunksToZarr(OUTPUT_PATH.value, template, output_chunks, num_threads=NUM_THREADS.value, storage_options=storage_options)
        )

    logging.info("Done")

if __name__ == "__main__":
    app.run(main)



"""This can be used to recreate the available pressure level tables if more are added"""
from ufs2arco.sources.rda_gfs_archive import RDAGFSArchive
from ufs2arco.sources.aws_hrrr_archive import AWSHRRRArchive
from ufs2arco.sources.aws_gefs_archive import AWSGEFSArchive

for name, Archive in zip(
    ["gfs", "gefs", "hrrr"],
    [RDAGFSArchive, AWSGEFSArchive, AWSHRRRArchive],
):

    kw = {}
    if name == "gefs":
        kw["member"]={"start": 0, "end": 1, "step": 1}


    levels = Archive(
        t0={"start": "2000-01-01", "end": "2000-01-02", "freq" :"6h"},
        fhr={"start": 0, "end": 6, "step": 6},
        **kw
    ).available_levels

    # Prepare rows with 3 columns each, padding as needed
    num_cols = 5
    rows = [levels[i:i+num_cols] for i in range(0, len(levels), num_cols)]

    with open(f"levels.{name}.rst", "w") as out:
        out.write(".. list-table:: Available Pressure Levels (hPa)\n")
        out.write(f"   :widths: {' '.join(['6']*num_cols)}\n")
        out.write("   :header-rows: 0\n\n")
        for row in rows:
            # Pad row to always have num_cols columns
            row_padded = tuple(row) + ('',) * (num_cols - len(row))
            out.write("   * - " + "\n     - ".join(str(cell) for cell in row_padded) + "\n")



import sys
import time

class Timer:
    """Simple timer class used to report runtime.
    """

    def __init__(self, filename=None):
        self._start_time = None
        self.filename = filename

    @property
    def is_running(self):
        """If this timer is running, return True

        Returns:
            is_running (bool): if this timer has been started, return True, otherwise False
        """
        return self._start_time is not None

    def start(self, mytitle=""):
        """Start a new timer and print what is about to be timed

        Args:
            mytitle (str, optional): indicate the thing that's about to run and be timed
        """
        if self.is_running:
            raise TimerError("Timer is running. Use .stop() to stop it")

        if mytitle != "":
            self._print(f" --- {mytitle} --- ")
        self._start_time = time.perf_counter()

    def get_elapsed(self):
        """Get time that has passed since timer started

        Returns:
            (float): amount of time elapsed in seconds
        """
        return time.perf_counter() - self._start_time

    def stop(self, mytitle="Elapsed time"):
        """Stop the timer, and report the elapsed time. Raise TimerError if
        no timer is running.

        Returns:
            (float): amount of time elapsed in seconds
        """
        if not self.is_running:
            raise TimerError("Timer is not running. Use .start() to start it")
        elapsed_time = self.get_elapsed()
        self._start_time = None
        if mytitle is not None:
            self._print(f"{mytitle}: {elapsed_time:.4f} seconds\n")
        return float(elapsed_time)

    def now(self):
        """Return the time elapsed since timer was started, without stopping the timer

        Returns:
            (float): elapsed time in seconds
        """
        return float(self.get_elapsed())

    def _print(self, *args, **kwargs):
        """Print the timing to :attr:`filename` if specified, or to screen.

        All arguments and keyword arguments are passed to ``print()``
        """

        if self.filename is None:
            print(*args, **kwargs)
        else:
            with open(self.filename,'a') as file:
                print(*args, file=file, **kwargs)

class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""



import logging
from typing import Optional
from copy import deepcopy
import re

import numpy as np
import pandas as pd
import xarray as xr
import zarr

from ufs2arco.sources import Source
from ufs2arco.targets import Target

logger = logging.getLogger("ufs2arco")

class Anemoi(Target):
    """
    Store dataset ready for anemoi

    Expected output has dimensions
        ``("time", "variable", "ensemble", "cell")``

    Use the rename argument to modify any of these before they get packed in the anemoi dataset.
    This might be useful if you want to train a model with the same variables, but from different datasets,
    so they have different names originally.

    Assumptions:
        * For :class:`EnsembleForecastSource` and :class:`ForecastSource` datasets, t0 gets renamed to time, and fhr is silently dropped
        * :attr:`do_flatten_grid` = ``True``
        * resolution = None, I have no idea where this gets set in anemoi-datasets
        * just setting use_level_index = False for now, but eventually it would be nice to use this flag to switch between how vertical level suffixes are labeled
        * if :attr:`sort_channels_by_level` is ``True``, then we'll make sure that channels go like variable_<level 0> -> variable_<largest level value>
    """

    # these should probably be options
    do_flatten_grid = True
    resolution = None
    use_level_index = False
    allow_nans = True
    data_dtype = np.float32

    # these are basically properties
    always_open_static_vars = True

    @property
    def sample_dims(self):
        if self._has_member:
            return ("time", "ensemble")
        else:
            return ("time",)

    @property
    def expanded_horizontal_dims(self):
        return tuple(self.protected_rename.get(d, d) for d in self.source.horizontal_dims)

    @property
    def horizontal_dims(self):
        if self.do_flatten_grid:
            return ("cell",)
        else:
            return self.expanded_horizontal_dims

    @property
    def datetime(self):
        if self._has_fhr:
            return self.source.t0 + pd.Timedelta(hours=self.source.fhr[0])
        else:
            return self.source.time

    @property
    def dates(self):
        return self.datetime

    @property
    def time(self):
        return np.arange(len(self.dates))

    @property
    def ensemble(self):
        if self._has_member:
            return self.source.member
        else:
            return [0]

    @property
    def start_date(self):
        return str(self.datetime[0]).replace(" ", "T")

    @property
    def end_date(self):
        return str(self.datetime[-1]).replace(" ", "T")

    @property
    def statistics_start_date(self):
        if self.statistics_period.get("start", None) is None:
            return self.start_date
        else:
            date = pd.Timestamp(self.statistics_period.get("start"))
            assert date in self.datetime, "{self.name}: could not find statistics_start_date within datetime"
            return str(date).replace(" ", "T")

    @property
    def statistics_end_date(self):
        if self.statistics_period.get("end", None) is None:
            return self.end_date
        else:
            date = pd.Timestamp(self.statistics_period.get("end"))
            assert date in self.datetime, "{self.name}: could not find statistics_end_date within datetime"
            return str(date).replace(" ", "T")

    @property
    def protected_rename(self) -> dict:
        protected_rename = {
            "latitude": "latitudes",
            "longitude": "longitudes",
        }
        if self._has_fhr:
            protected_rename["valid_time"] = "dates"
        else:
            protected_rename["time"] = "dates"

        if self._has_member:
            protected_rename["member"] = "ensemble"
        return protected_rename


    @property
    def dim_order(self):
        return ("time", "variable", "ensemble") + self.horizontal_dims


    def __init__(
        self,
        source: Source,
        chunks: dict,
        store_path: str,
        rename: Optional[dict] = None,
        forcings: Optional[tuple | list] = None,
        statistics_period: Optional[dict] = None,
        compute_temporal_residual_statistics: Optional[bool] = False,
        sort_channels_by_levels: Optional[bool] = False,
    ) -> None:

        super().__init__(
            source=source,
            chunks=chunks,
            store_path=store_path,
            rename=rename,
            forcings=forcings,
            statistics_period=statistics_period,
            compute_temporal_residual_statistics=compute_temporal_residual_statistics,
        )

        self.sort_channels_by_levels = sort_channels_by_levels
        # additional checks
        if self._has_fhr:
            assert len(self.source.fhr) == 1, \
                f"{self.name}.__init__: Can only use this class with len(fhr)==1, no multiple lead times"

        renamekeys = list(self.rename.keys())
        protected = list(self.protected_rename.keys()) + list(self.protected_rename.values())
        for key in renamekeys:
            if key in protected or self.rename[key] in protected:
                logger.info(f"{self.name}.__init__: can't rename {key} -> {self.rename[key]}, either key or val is in a protected list. I'll drop it and forget about it.")
                self.rename.pop(key)


    def get_expanded_dim_order(self, xds):
        """this is used in :meth:`map_static_to_expanded`"""
        return ("time", "ensemble") + tuple(xds.attrs["stack_order"])


    def apply_transforms_to_sample(
        self,
        xds: xr.Dataset,
    ) -> xr.Dataset:

        if self._has_fhr:
            xds = xds.squeeze("fhr", drop=True)
            xds = xds.swap_dims({"t0": "valid_time"})

        if not self._has_member:
            xds = xds.expand_dims({"ensemble": self.ensemble})

        xds = super().apply_transforms_to_sample(xds)
        xds = self._map_datetime_to_index(xds)
        xds = self._map_levels_to_suffixes(xds)
        xds = self._map_static_to_expanded(xds)
        xds = xds.transpose(* (("time", "ensemble") + tuple(xds.attrs["stack_order"])) )
        xds = self._stackit(xds)
        xds = self._calc_sample_stats(xds)
        if self.do_flatten_grid:
            xds = self._flatten_grid(xds)
        xds = xds.transpose(*self.dim_order)
        xds = xds.reset_coords()
        xds = xds[sorted(xds.data_vars)]
        return xds


    def manage_coords(self, xds: xr.Dataset) -> xr.Dataset:
        attrs = {
            "allow_nans": self.allow_nans,
            "ensemble_dimension": len(self.ensemble),
            "flatten_grid": self.do_flatten_grid,
            "resolution": str(self.resolution),
            "start_date": self.start_date,
            "end_date": self.end_date,
            "frequency": self.datetime.freqstr,
            "statistics_start_date": self.statistics_start_date,
            "statistics_end_date": self.statistics_end_date,
        }
        xds.attrs.update(attrs)
        return xds


    def rename_dataset(self, xds: xr.Dataset) -> xr.Dataset:
        """
        In addition to any user specified renamings...
        This takes the default source dimensions and renames them to the default anemoi dimensions:

        (t0, member, level, latitude, longitude) -> (dates, ensemble, level, latitudes, longitudes)
        or
        (time, level, latitude, longitude) -> (dates, ensemble, level, latitudes, longitudes)

        Args:
            xds (xr.Dataset): a dataset directly from the source

        Returns:
            xds (xr.Dataset): with renaming as above
        """
        # first, rename the protected list
        # note that we do not update the attribute rename with protected_rename
        # key: values because of the way time is handled ...
        # it is convenient to rename time to dates
        # however... anemoi target has "time" as a logical index in its sample_dims
        # which is completely different...
        # because datamover.create_container relies on Target.renamed_sample_dims attribute,
        # this would rename time to dates, and then create a container with the dates as a dimension...
        # even though it's not
        # then, add on to that the fact that we drop the dates vector and recreate it at the very end,
        # due to the differences in the way xarray and zarr handle datetime objects,
        # we really have to treat these protected quantities differently
        xds = xds.rename(self.protected_rename)
        xds = super().rename_dataset(xds)
        return xds


    def _map_datetime_to_index(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Turn the datetime vector into a logical index and swap the dimensions

        (dates, ensemble, level, latitudes, longitudes) -> (time, ensemble, level, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with time dimension "dates"

        Returns:
            xds (xr.Dataset): with new time dimension "time" ("dates" is still there)
        """

        t = [list(self.datetime).index(date) for date in xds["dates"].values]
        xds["time"] = xr.DataArray(
            t,
            coords=xds["dates"].coords,
            dims=xds["dates"].dims,
            attrs={
                "description": "logical time index",
            },
        )
        xds = xds.swap_dims({"dates": "time"})

        # anemoi needs "dates" to be stored as a specific dtype
        # it turns out that this is hard to do consistently with xarray and zarr
        # especially with this "write container" + "fill incrementally" workflow
        # so... let's just store "dates" during aggregate_stats
        xds = xds.drop_vars("dates")
        return xds

    def _map_levels_to_suffixes(self, xds):
        """
        Take each of the 3D variables, and make them n_levels x 2D variables with the suffix _{level} for each level

        (time, ensemble, level, latitudes, longitudes) -> (time, ensemble, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with all variables, maybe 2D maybe 3D

        Returns:
            nds (xr.Dataset): 3D variables expanded into n_level x 2D variables
        """

        nds = xr.Dataset()
        nds.attrs["variables_metadata"] = dict()

        for name in xds.data_vars:
            meta = {
                "mars": {
                    "date": str(self.datetime[xds.time.values[0]]).replace("-","")[:8],
                    "param": name,
                    "step": 0, # this is the fhr=0 assumption
                    "time": str(self.datetime[xds.time.values[0]]).replace("-","").replace(" ", "").replace(":","")[8:12], # no idea what this should be actually
                    "valid_datetime": str(self.datetime[xds.time.values[0]]).replace(" ", "T"),
                    "variable": name,
                },
            }
            if len(self.ensemble) > 1:
                meta["mars"]["number"] = list(self.ensemble)
            if "level" in xds[name].dims:
                for level in xds[name].level.values:
                    idx = self._get_level_index(xds, level)
                    ilevel = int(level)
                    ilevel = ilevel if ilevel == level else level
                    suffix_name = f"{name}_{ilevel}" if not self.use_level_index else f"{name}_{idx}"
                    nds[suffix_name] = xds[name].sel({"level": level}, drop=True)
                    units = xds["level"].attrs.get('units', '')
                    nds[suffix_name].attrs.update(
                        {
                            "level": ilevel,
                            "level_description": f"{name} at vertical level (index, value) = ({idx}, {ilevel}{units})",
                            "level_index": idx,
                        },
                    )
                    nds.attrs["variables_metadata"][suffix_name] = deepcopy(meta)
                    nds.attrs["variables_metadata"][suffix_name]["mars"]["level"] = ilevel if not self.use_level_index else idx

                if "remapping" not in nds.attrs:
                    nds.attrs["remapping"] = {"param_level": "{param}_{levelist}"}
                else:
                    if "param_level" not in nds.attrs["remapping"].keys():
                        nds.attrs["remapping"]["param_level"] = "{param}_{levelist}"
            else:
                nds[name] = xds[name]

                if "computed_forcing" in xds[name].attrs and "constant_in_time" in xds[name].attrs:
                    nds.attrs["variables_metadata"][name] = {
                        "computed_forcing": xds[name].attrs["computed_forcing"],
                        "constant_in_time": xds[name].attrs["constant_in_time"],
                    }
                else:
                    nds.attrs["variables_metadata"][name] = deepcopy(meta)
                # Is attributes here a hack? Add the "field_shape" here
                # so that it's in the order of the data arrays, not in the dataset order
                # (they could be different)
                if "field_shape" not in nds.attrs and all(d in xds[name].dims for d in self.expanded_horizontal_dims):
                    stack_order = list(d for d in xds[name].dims if d in self.expanded_horizontal_dims)
                    nds.attrs["stack_order"] = stack_order
                    nds.attrs["field_shape"] = list(len(xds[d]) for d in stack_order)

        return nds


    @staticmethod
    def _get_level_index(xds: xr.Dataset, value: int | float) -> int | float:
        return xds["level"].values.tolist().index(value)


    def _sort_channels_by_levels(self, item: tuple) -> tuple:
        """
        If we have a dataset with e.g. geopotential (gh) at 100, 150, 200, ... 1000 hPa
        then the :meth:`_map_levels_to_suffixes` method returns them like this:

            ["gh_100", "gh_1000", "gh_150", ...]

        this method will sort the list of data_vars so they are like this:

            ["gh_100", "gh_150", ... "gh_1000"]

        This is used internally as a key for the sorted function in :meth:`stackit`
        """
        # Match variable names with an underscore followed by a number
        match = re.match(r"(.+)_(\d+)$", item)
        if match:
            var_name, num = match.groups()
            return (var_name, int(num))  # Sort by name, then numeric value

        # Ensure variables like 't2m' are grouped correctly
        return (item, -1)  # Non-numeric suffix variables come before numbered ones


    def _map_static_to_expanded(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Take each variable that does not have any of the (time, ensemble, latitudes, longitudes),
        and expand it so they all look the same

        Args:
            xds (xr.Dataset): with some static variable maybe

        Returns:
            xds (xr.Dataset): all data_vars have the same shape
        """

        for key in xds.data_vars:
            for d in xds.dims:
                if d not in xds[key].dims:
                    xds[key] = xds[key].expand_dims({d: xds[d]})

        return xds


    def _stackit(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Stack the multivariate dataset to a single data array with all variables (and vertical levels) stacked together

        (time, ensemble, latitudes, longitudes) -> (time, ensemble, variable, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with all variables, 3D variables as multiple 2D variables with suffixes

        Returns:
            xds (xr.Dataset): with "data" DataArray, which has all variables/levels stacked together
        """
        varlist = sorted(
            list(xds.data_vars),
            key=self._sort_channels_by_levels if self.sort_channels_by_levels else None,
        )
        channel = [i for i, _ in enumerate(varlist)]
        channel = xr.DataArray(
            channel,
            coords={"variable": channel},
            dims="variable",
        )

        # this might be nice, but it doesn't exist in anemoi
        # and it causes problems with the container / fill workflow
        # it should just be added as a coordinate... but again it's not in anemoi
        #channel_names = xr.DataArray(
        #    varlist,
        #    coords=channel.coords,
        #    dims=channel.dims,
        #)

        data_vars = xr.concat(
            [
                xds[name].expand_dims(
                    {"variable": [this_channel]},
                ).astype(
                    self.data_dtype,
                )
                for this_channel, name in zip(channel, varlist)
            ],
            dim="variable",
            combine_attrs="drop",
        )
        nds = data_vars.to_dataset(name="data")
        nds.attrs = xds.attrs.copy()
        # not making this a data array, even though it might be kinda nice
        nds.attrs["variables"] = varlist
        return nds

    def _flatten_grid(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Flatten (latitudes, longitudes) -> (cell,)

        (time, ensemble, variable, latitudes, longitudes) -> (time, ensemble, variable, cell)

        Args:
            xds (xr.Dataset): with expanded grid

        Returns:
            xds (xr.Dataset): with grid flattened to "cell"
        """
        nds = xds.stack(cell2d=xds.attrs["stack_order"])
        nds["cell"] = xr.DataArray(
            np.arange(len(nds["cell2d"])),
            coords=nds["cell2d"].coords,
            dims=nds["cell2d"].dims,
            attrs={
                "description": f"logical index for 'cell2d', which is a flattened lon x lat array",
            },
        )
        nds = nds.swap_dims({"cell2d": "cell"})

        # For some reason, there's a failure when trying to store this multi-index
        # it's not needed in Anemoi, so no need to keep it anyway.
        nds = nds.drop_vars("cell2d")
        return nds

    def _calc_sample_stats(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Compute statistics for this data sample, which will be aggregated later

        Args:
            xds (xr.Dataset): with just the data and coordinates

        Returns:
            xds (xr.Dataset): with the following statistics, each with an "_array" suffix,
                in order to indicate that the result will still have "time" and "ensemble" dimensions
                that will need to get aggregated
                ["count", "has_nans", "maximum", "minimum", "squares", "sums"]
        """

        dims = list(self.expanded_horizontal_dims)
        xds["count_array"] = (~np.isnan(xds["data"])).sum(dims, skipna=self.allow_nans).astype(np.float64)
        xds["has_nans_array"] = np.isnan(xds["data"]).any(dims)
        xds["maximum_array"] = xds["data"].max(dims, skipna=self.allow_nans).astype(np.float64)
        xds["minimum_array"] = xds["data"].min(dims, skipna=self.allow_nans).astype(np.float64)
        xds["squares_array"] = (xds["data"]**2).sum(dims, skipna=self.allow_nans).astype(np.float64)
        xds["sums_array"] = xds["data"].sum(dims, skipna=self.allow_nans).astype(np.float64)
        return xds


    def finalize(self, topo) -> None:
        """Finalize the dataset with
        * dates
        * stats
        * temporal stats (if specified)
        """

        self.add_dates(topo)

        logger.info(f"Aggregating statistics")
        self.aggregate_stats(topo)
        logger.info(f"Done aggregating statistics\n")

        if self.compute_temporal_residual_statistics:
            logger.info(f"Computing temporal residual statistics")
            self.calc_temporal_residual_stats(topo)
            logger.info(f"Done computing temporal residual statistics\n")

        super().finalize(topo=topo)


    def add_dates(self, topo) -> None:
        """Deal with the dates issue

        for some reason, it is a challenge to get the datetime64 dtype to open
        consistently between zarr and xarray, and
        it is much easier to deal with this all at once here
        than in the create_container and incrementally fill workflow.
        """

        if topo.is_root:
            xds = xr.open_zarr(self.store_path)
            attrs = xds.attrs.copy()

            nds = xr.Dataset()
            nds["dates"] = xr.DataArray(
                self.datetime,
                coords=xds["time"].coords,
            )
            nds["dates"].encoding = {
                "dtype": "datetime64[s]",
                "units": "seconds since 1970-01-01",
            }

            # store it, first copying the attributes over
            nds.attrs = attrs
            nds.to_zarr(self.store_path, mode="a")
            logger.info(f"{self.name}.add_dates: dates appended to the dataset\n")


    def aggregate_stats(self, topo) -> None:
        """Aggregate statistics over "time" and "ensemble" dimension...
        I'm assuming that this is relatively inexpensive without the spatial dimension

        This will store an array with the statistics

            ["count", "has_nans", "maximum", "mean", "minimum", "squares", "stdev", "sums"]

        and it will get rid of the "_array" versions of the statistics
        """

        xds = xr.open_zarr(self.store_path)
        attrs = xds.attrs.copy()

        # get the start/end times for computing statistics
        # in terms of logical time index values
        start_idx = list(self.datetime).index(pd.Timestamp(self.statistics_start_date))
        end_idx = list(self.datetime).index(pd.Timestamp(self.statistics_end_date))
        xds = xds.sel(time=slice(start_idx, end_idx))

        dims = ["time", "ensemble"]
        time_indices = np.array_split(np.arange(len(xds["time"])), topo.size)
        local_indices = time_indices[topo.rank]

        vidx = xds["variable"].values
        count = np.zeros_like(vidx, dtype=xds["count_array"].dtype)
        has_nans = np.full_like(vidx, fill_value=False, dtype=xds["has_nans_array"].dtype)
        maximum = np.full_like(vidx, fill_value=-np.inf, dtype=xds["maximum_array"].dtype)
        minimum = np.full_like(vidx, fill_value=np.inf, dtype=xds["minimum_array"].dtype)
        squares = np.zeros_like(vidx, dtype=xds["squares_array"].dtype)
        sums = np.zeros_like(vidx, dtype=xds["sums_array"].dtype)

        logger.info(f"{self.name}.aggregate_stats: Performing local computations")
        if local_indices.size > 0:

            lds = xds.isel(time=local_indices)
            local_count = lds["count_array"].sum(dims).compute().values
            local_has_nans = lds["has_nans_array"].any(dims).compute().values
            local_maximum = lds["maximum_array"].max(dims).compute().values
            local_minimum = lds["minimum_array"].min(dims).compute().values
            local_squares = lds["squares_array"].sum(dims).compute().values
            local_sums = lds["sums_array"].sum(dims).compute().values

        else:

            local_count = count.copy()
            local_has_nans = has_nans.copy()
            local_maximum = maximum.copy()
            local_minimum = minimum.copy()
            local_squares = squares.copy()
            local_sums = sums.copy()


        # reduce results
        logger.info(f"{self.name}.aggregate_stats: Communicating results to root")
        topo.sum(local_count, count)
        topo.any(local_has_nans, has_nans)
        topo.max(local_maximum, maximum)
        topo.min(local_minimum, minimum)
        topo.sum(local_squares, squares)
        topo.sum(local_sums, sums)
        logger.info(f"{self.name}.aggregate_stats: ... done communicating")

        # the rest is done on the root rank
        if topo.is_root:
            nds = xr.Dataset()
            kw = {"coords": xds["variable"].coords}
            nds["count"] = xr.DataArray(count, **kw)
            nds["has_nans"] = xr.DataArray(has_nans, **kw)
            nds["maximum"] = xr.DataArray(maximum, **kw)
            nds["minimum"] = xr.DataArray(minimum, **kw)
            nds["squares"] = xr.DataArray(squares, **kw)
            nds["sums"] = xr.DataArray(sums, **kw)

            # now add mean & stdev
            nds["mean"] = nds["sums"] / nds["count"]
            variance = nds["squares"] / nds["count"] - nds["mean"]**2
            nds["stdev"] = xr.where(variance >= 0, np.sqrt(variance), 0.)

            # store it, first copying the attributes over
            nds.attrs = attrs
            nds.to_zarr(self.store_path, mode="a")
            logger.info(f"{self.name}.aggregate_stats: Stored aggregated stats")

        # unclear if this barrier is necessary...
        topo.barrier()


    def calc_temporal_residual_stats(self, topo):

        xds = xr.open_zarr(self.store_path)
        attrs = xds.attrs.copy()

        # get the start/end times for computing statistics
        # in terms of logical time index values
        start_idx = list(self.datetime).index(pd.Timestamp(self.statistics_start_date))
        end_idx = list(self.datetime).index(pd.Timestamp(self.statistics_end_date))
        xds = xds.sel(time=slice(start_idx, end_idx))

        stdev = xds["stdev"].load()
        mean = xds["mean"].load()

        data_norm = (xds["data"] - mean)/stdev
        data_diff = data_norm.diff("time")
        n_time = len(data_diff["time"])
        time_indices = np.array_split(np.arange(n_time), topo.size)
        local_indices = time_indices[topo.rank]

        residual_variance = np.zeros_like(xds["variable"].values, dtype=np.float64)

        logger.info(f"{self.name}.calc_temporal_residual_stats: Performing local computations")
        if local_indices.size > 0:
            mdims = [d for d in xds["data"].dims if d not in ("variable", "time")]
            local_data_diff = data_diff.isel(time=local_indices).astype(np.float64)
            local_residual_variance = (local_data_diff**2).mean(mdims).sum("time").compute().values
            local_residual_variance /= n_time
        else:
            local_residual_variance = residual_variance.copy()

        logger.info(f"{self.name}.calc_temporal_residual_stats: Communicating results to root")
        topo.sum(local_residual_variance, residual_variance)

        logger.info(f"{self.name}.calc_temporal_residual_stats: ... done communicating")

        if topo.is_root:
            nds = xr.Dataset()
            nds.attrs = attrs
            nds["residual_stdev"] = xr.DataArray(np.sqrt(residual_variance), coords=xds["variable"].coords)

            # compute geomtric mean by log-exp trick (since scipy isn't a dependency to ufs2arco)
            # ignore 0 values, since this is probably from static variables
            rstdev = nds["residual_stdev"].where(nds["residual_stdev"] > 0)
            denominator = np.exp(np.log(rstdev).mean("variable").values)
            nds["gmean_residual_stdev"] = nds["residual_stdev"] / denominator

            nds.to_zarr(self.store_path, mode="a")
            logger.info(f"{self.name}.calc_temporal_residual_stats: Stored temporal residual stats")

        # unclear if this barrier is necessary
        topo.barrier()


    def handle_missing_data(self, missing_data: list[dict]) -> None:
        """Take a list of dicts, with dimensions of missing data. Keep track of any missing dates
        (via "t0" or "time") and store that in the zarr.

        Note that anemoi doesn't really care if we have 30 ensemble members that are good to go for a missing date,
        if one is missing they all are.

        Note: it is assumed this is only called from the root process

        Args:
            missing_data (list[dict]): list with missing data dicts
        """

        missing_dates = []
        zds = zarr.open(self.store_path, mode="a")

        for missing_sample in missing_data:
            if self._has_fhr:
                valid_time = pd.Timestamp(missing_sample["t0"]) + pd.Timedelta(hours=missing_sample["fhr"])
                missing_dates.append(str(valid_time))
            else:
                missing_dates.append(missing_sample["time"])

        zds.attrs["missing_dates"] = missing_dates
        zarr.consolidate_metadata(self.store_path)



import os
import yaml
import warnings
import xarray as xr
import numpy as np

from .ufsregridder import UFSRegridder

try:
    import xesmf as xe
except ImportError:
    pass


class CICE6Regridder(UFSRegridder):
    """
    Regrid cice dataset that is on a tripolar grid to a different grid (primarily Gaussian grid).


    Optional fields in config:
        rotation_file (str): path to file containing rotation fields "sin_rot" and "cos_rot", required for regridding vector fields
        weights_file_t2t (str): path to t2t interpolation weights file
        weights_file_u2t (str): path to u2t interpolation weights file
        periodic (bool): Is the grid periodic in longitude?
    """

    __doc__ = __doc__ + UFSRegridder.__doc__

    rg_tt = None
    rg_ut = None

    def __init__(
        self,
        lats1d_out: np.array,
        lons1d_out: np.array,
        ds_in: xr.Dataset,
        config_filename: str,
    ) -> None:
        super(CICE6Regridder, self).__init__(
            lats1d_out, lons1d_out, ds_in, config_filename
        )

    def _create_regridder(self, ds_in: xr.Dataset) -> None:

        # create rotation dataset
        self.rotation_file = self.config.get("rotation_file", None)
        self.ds_rot = None
        if self.rotation_file is not None:
            ds_rot = xr.open_dataset(self.rotation_file)
            self.ds_rot = xr.Dataset()
            self.ds_rot["cos_rot"] = np.cos(ds_rot["ANGLE"])
            self.ds_rot["sin_rot"] = np.sin(ds_rot["ANGLE"])
        elif "ANGLE" in ds_in:
            self.ds_rot = xr.Dataset()
            self.ds_rot["cos_rot"] = np.cos(ds_in["ANGLE"])
            self.ds_rot["sin_rot"] = np.sin(ds_in["ANGLE"])
        else:
            warnings.warn(
                f"CICE6Regridder._create_regridder: Could not find 'rotation_file' in configuration yaml. "
                f"Vector fields will be silently ignored."
            )

        # create input dataset with t-/u- coordinates
        ds_in_t = xr.Dataset()
        ds_in_t["lon"] = ds_in["TLON"]
        ds_in_t["lat"] = ds_in["TLAT"]
        ds_in_u = xr.Dataset()
        ds_in_u["lon"] = ds_in["ULON"]
        ds_in_u["lat"] = ds_in["ULAT"]

        # create output dataset
        lons, lats = np.meshgrid(self.lons1d_out, self.lats1d_out)
        grid_out = xr.Dataset()
        grid_out["lon"] = xr.DataArray(lons, dims=["lat", "lon"])
        grid_out["lat"] = xr.DataArray(lats, dims=["lat", "lon"])

        # get nlon/nlat for input/output datsets
        nlon_i = ds_in.sizes["ni"]
        nlat_i = ds_in.sizes["nj"]
        nlon_o = len(self.lons1d_out)
        nlat_o = len(self.lats1d_out)
        self.ires = f"{nlon_i}x{nlat_i}"
        self.ores = f"{nlon_o}x{nlat_o}"

        # paths to interpolation weights files
        wfiles = dict()
        interp_method = self.config["interp_method"]
        for key in ["weights_file_t2t", "weights_file_u2t"]:
            vin = key[-3]
            default = f"weights-cice6-{self.ires}.C{vin}.{self.ores}.Ct.{interp_method}.nc"
            path = self.config.get(key, None)
            wfiles[key] = path if path is not None else default

        # create regridding instances
        periodic = self.config["periodic"]
        reuse = os.path.exists(wfiles["weights_file_t2t"])
        self.rg_tt = xe.Regridder(
            ds_in_t,
            grid_out,
            interp_method,
            periodic=periodic,
            reuse_weights=reuse,
            filename=wfiles["weights_file_t2t"],
        )
        if self.ds_rot is not None:
            reuse = os.path.exists(wfiles["weights_file_u2t"])
            self.rg_ut = xe.Regridder(
                ds_in_u,
                ds_in_t,
                interp_method,
                periodic=periodic,
                reuse_weights=reuse,
                filename=wfiles["weights_file_u2t"],
            )

    def regrid(self, ds_in: xr.Dataset) -> xr.Dataset:
        ds_out = []

        # CICE6 dataset specific variable and coordinate names
        coords_xy = {"nj", "ni"}
        variable_map = {
            "uvel": ("vvel", "U"),
            "vvel": (None, "skip"),
            "strairx": ("strairy", "U"),
            "strairy": (None, "skip"),
            "strocnx": ("strocny", "U"),
            "strocny": (None, "skip"),
            "uvel_d": ("vvel_d", "U"),
            "vvel_d": (None, "skip"),
            "strairx_d": ("strairy_d", "U"),
            "strairy_d": (None, "skip"),
            "strocnx_d": ("strocny_d", "U"),
            "strocny_d": (None, "skip"),
            "uvel_h": ("vvel_h", "U"),
            "vvel_h": (None, "skip"),
            "strairx_h": ("strairy_h", "U"),
            "strairy_h": (None, "skip"),
            "strocnx_h": ("strocny_h", "U"),
            "strocny_h": (None, "skip"),
        }

        return super(CICE6Regridder, self).regrid_tripolar(
            ds_in,
            self.ds_rot,
            self.rg_tt,
            self.rg_ut,
            self.rg_ut,
            coords_xy,
            variable_map,
        )



import yaml
import numpy as np
import xarray as xr
from abc import ABC, abstractmethod
import warnings

from .gaussian_grid import gaussian_latitudes

try:
    from xesmf import Regridder

    _has_xesmf = True
except ImportError:
    Regridder = None
    _has_xesmf = False


class UFSRegridder(ABC):
    """

    Args:
        lats1d_out (array_like): One-dimensional array containing latitudes of the output grid.
        lons1d_out (array_like): One-dimensional array containing longitudes of the output grid.
        config_filename (str): YAML config file specifying options for regridding.
        ds_in (xarray.Dataset): CICE6 xarray Dataset
        interp_method (str, optional): Type of interpolation, default="bilinear".

    Example:

        This example uses the :class:`CICE6Regridder`, but similar capabilities exist for :class:`MOM6Regridder`.
        Construct the output grid. The following static methods are provided in this class
            - compute_gaussian_grid
            - compute_latlon_grid
            - read_grid

        >>> lats, lons = CICE6Regridder.compute_gaussian_grid(180, 360)

        Open dataset using xarray or preferably :meth:`CICE6Dataset.open_dataset`

        >>> ds = xr.open_mfdataset("./input/{ocn_,iceh_}*.nc")

        Construct :class:`CICE6Regridder` object, specifying output grid lats & lons, dataset and config file

        >>> rg = CICE6Regridder(lats, lons, ds, config_filename = "config-replay.yaml")

        Regrid dataset

        >>> ds = rg.regrid(ds)
    """

    def __init__(
        self,
        lats1d_out: np.array,
        lons1d_out: np.array,
        ds_in: xr.Dataset,
        config_filename: str,
    ) -> None:
        """
        Initialize the UFSRegridder object.
        """

        if not _has_xesmf:
            raise ImportError(
                f"Cannot import xesmf. Install with 'conda install -c conda-forge xesmf',"
                f" and note the environment must be deactivated and reactivated"
            )

        super(UFSRegridder, self).__init__()
        name = self.__class__.__name__
        # Look for the "regrid" section of e.g. MOM6Dataset
        name = name.replace("Regridder", "Dataset")

        # Load configuration from YAML file
        with open(config_filename, "r") as f:
            dataset_config = yaml.safe_load(f).get(name, {})
            self.config = dataset_config.get("regrid", {})
            self.data_vars = dataset_config.get("data_vars", "")

        # specify an output resolution
        self.lats1d_out = lats1d_out
        self.lons1d_out = lons1d_out
        self.interp_method = self.config["interp_method"]

        # create regridder
        self._create_regridder(ds_in)

    @abstractmethod
    def _create_regridder(self, ds_in: xr.Dataset) -> None:
        """
        Create regridder: computes weights and creates three xesmf.Regridder instances.
        Args:
            ds_in (xarray.Dataset): xarray dataset needed for creating the regridder.
        """

    @staticmethod
    def compute_gaussian_grid(nlat, nlon) -> (np.array, np.array):
        """Compute gaussian grid latitudes and longitudes"""
        latitudes, _ = gaussian_latitudes(nlat // 2)
        longitudes = np.linspace(0, 360, nlon, endpoint=False)
        return latitudes, longitudes

    @staticmethod
    def compute_latlon_grid(nlat, nlon) -> (np.array, np.array):
        """Compute regular latlong grid coordinates"""
        latitudes = np.linspace(-90, 90, nlat, endpoint=False)
        longitudes = np.linspace(0, 360, nlon, endpoint=False)
        return latitudes, longitudes

    @staticmethod
    def read_grid(
        file, lats_name="grid_yt", lons_name="grid_xt"
    ) -> (np.array, np.array):
        """Read grid latitudes and longitudes from a netcdf file"""
        ds = xr.open_dataset(file)
        latitudes = ds[lats_name].values
        longitudes = ds[lons_name].values
        ds.close()
        return latitudes, longitudes

    def regrid_tripolar(
        self,
        ds_in: xr.Dataset,
        ds_rot: xr.Dataset,
        rg_tt: Regridder,
        rg_ut: Regridder,
        rg_vt: Regridder,
        coords_xy: set,
        variable_map: dict,
    ) -> xr.Dataset:
        """
        Regrid an xarray dataset.

        Args:
            ds_in (xarray.Dataset): Input xarray Dataset to regrid.
            ds_rot (xarray.Dataset): Dataset containing rotation angles
            rg_tt (xesmf.Regridder): Regridder object mapping t-points to output grid
            rg_ut (xesmf.Regridder): Regridder object mapping u-points to t-points
            rg_vt (xesmf.Regridder): Regridder object mapping v-points to t-points
            coords_xy (set): A list of spatial coordinate names used in ds_in
            variable_map (dict): A dictionary mapping vector field names to a set e.g. sea-surface velocity
                       "SSU": ("SSV", "U"),
                       "SSV": (None, "skip)

        Returns:
            ds_out (xarray.Dataset): Regridded xarray Dataset
        """
        ds_out = []

        # Only loop through data variables we want to keep
        # If user didn't give any, then loop through all variables
        data_vars = self.data_vars if len(self.data_vars) > 0 else list(ds_in.keys())

        # make sure desired variables are actually in the dataset
        data_vars = [k for k in data_vars if k in ds_in]

        # interate through variables and regrid
        for var in data_vars:
            coords = ds_in[var].coords.to_index()

            # must have lat/lon like coordinates, otherwise append copy
            if len(set(coords.names) & coords_xy) < 2:
                ds_out.append(ds_in[var])
                continue

            # choose regridding type
            if var in variable_map.keys():
                var2, pos = variable_map[var]
            else:
                var2, pos = (None, "T")

            # scalar fields
            if pos == "T":
                # interpolate
                da_out = rg_tt(ds_in[var])
                ds_out.append(_xda_to_xds(da_out, var, ds_in[var].attrs))

            # vector fields
            if pos == "U" and ds_rot is not None:
                # interpolate to t-points
                interp_u = rg_ut(ds_in[var])
                interp_v = rg_vt(ds_in[var2])

                # rotate to earth-relative
                urot = (
                    interp_u * ds_rot.cos_rot + interp_v * ds_rot.sin_rot
                )
                vrot = (
                    interp_v * ds_rot.cos_rot - interp_u * ds_rot.sin_rot
                )

                # interoplate
                uinterp_out = rg_tt(urot)
                vinterp_out = rg_tt(vrot)

                # append vars
                ds_out.append(_xda_to_xds(uinterp_out, var, ds_in[var].attrs))
                ds_out.append(_xda_to_xds(vinterp_out, var2, ds_in[var2].attrs))

            elif pos == "U" and ds_rot is None:
                warnings.warn(
                    f"UFSRegridder.regrid_tripolar: rotation_file not provided, skipping vector fields ({var}, {var2})"
                )

        # merge dataarrays into a dataset and set attributes
        ds_out = xr.merge(ds_out, compat="override")
        ds_out.attrs = ds_in.attrs
        ds_out = ds_out.assign_coords(lon=("lon", self.lons1d_out))
        ds_out = ds_out.assign_coords(lat=("lat", self.lats1d_out))
        ds_out = ds_out.assign_coords(time=("time", ds_in.time.values))
        ds_out["lon"].attrs.update(
            {
                "units": "degrees_east",
                "axis": "X",
                "standard_name": "longitude",
            }
        )
        ds_out["lat"].attrs.update(
            {
                "units": "degrees_north",
                "axis": "Y",
                "standard_name": "latiitude",
            }
        )

        return ds_out


def _xda_to_xds(da_out, var, attrs):
    """helper to update attrs and prepare for ds_out"""
    xds = da_out.to_dataset(name=var)
    xds[var].attrs.update(
        {
            "long_name": attrs["long_name"],
            "units": attrs["units"],
        }
    )
    return xds



import os
import logging
import warnings
from typing import Optional, Any, List

try:
    from mpi4py import MPI
    _has_mpi = True
except ImportError:
    _has_mpi = False
    warnings.warn(f"ufs2arco.mpi: Unable to import mpi4py, cannot use this module")

from .log import SimpleFormatter

logger = logging.getLogger("ufs2arco")

class MPITopology:
    """
    Handles MPI-based parallel processing topology.
    """

    def __init__(
        self,
        log_dir: Optional[str] = None,
        log_level: int = logging.INFO,
    ) -> None:
        """
        Initializes the MPI topology.

        Args:
            log_dir (Optional[str]): Directory for storing logs.
            log_level (int): Logging level.
        """

        assert _has_mpi, f"MPITopology.__init__: Unable to import mpi4py, cannot use this class"

        self.required_level = MPI.THREAD_MULTIPLE
        self.provided_level = MPI.Query_thread()
        self.comm = MPI.COMM_WORLD
        self.rank = self.comm.Get_rank()
        self.size = self.comm.Get_size()
        self.root = 0
        self.pid = os.getpid()
        self.friends = tuple(ii for ii in range(self.size) if ii!=self.root)

        self._init_log(log_dir=log_dir, level=log_level)
        logger.info(str(self))


    def __str__(self):
        msg = f"\nMPITopology Summary\n" +\
            f"-------------------\n" +\
            f"comm: {self.comm.Get_name()}\n"
        for key in ["rank", "size"]:
            msg += f"{key:<18s}: {getattr(self, key):02d}\n"
        msg += f"{'pid':<18s}: {self.pid}\n"
        msg += f"{'log_dir':<18s}: {self.log_dir}\n"
        msg += f"{'logfile':<18s}: {self.logfile}\n"
        msg += "Thread Support\n"+\
            f"{'required_level':<18s}: {self.required_level}\n" +\
            f"{'provided_level':<18s}: {self.provided_level}\n"
        return msg


    def _init_log(self, log_dir, level=logging.INFO):
        self.log_dir = "./" if log_dir is None else log_dir
        if self.is_root:
            if not os.path.isdir(self.log_dir):
                os.makedirs(self.log_dir)
        self.comm.Barrier()
        self.logfile = f"{self.log_dir}/log.{self.rank:04d}.{self.size:04d}.out"

        logger.setLevel(level=level)
        formatter = SimpleFormatter(fmt="[%(relativeCreated)d s] [%(levelname)-7s] %(message)s")
        handler = logging.FileHandler(self.logfile, mode="w+")
        handler.setLevel(level=level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)


    @property
    def is_root(self) -> bool:
        """
        Checks if the current process is the root process.

        Returns:
            bool: True if the current process is the root, False otherwise.
        """
        return self.rank == self.root


    def barrier(self) -> None:
        """
        Synchronizes all processes at a barrier.
        """
        self.comm.Barrier()

    def gather(self, local_array):
        return self.comm.gather(local_array, root=self.root)

    def Reduce(self, local_array, result_buffer, op):
        self.comm.Reduce(local_array, result_buffer, op=op, root=self.root)

    def sum(self, local_array, result_buffer):
        self.Reduce(local_array, result_buffer, op=MPI.SUM)

    def max(self, local_array, result_buffer):
        self.Reduce(local_array, result_buffer, op=MPI.MAX)

    def min(self, local_array, result_buffer):
        self.Reduce(local_array, result_buffer, op=MPI.MIN)

    def any(self, local_array, result_buffer):
        self.Reduce(local_array, result_buffer, op=MPI.LOR)


class SerialTopology:
    """
    Looks and acts like the MPITopology, but works in cases when mpi is not importable.

    So... basically it's a logger
    """
    def __init__(
        self,
        log_dir: Optional[str] = None,
        log_level: int = logging.INFO,
    ) -> None:
        """
        Args:
            log_dir (Optional[str]): Directory for storing logs.
            log_level (int): Logging level.
        """

        self.root = 0
        self.pid = os.getpid()
        self.size = 1
        self.rank = 0

        self._init_log(log_dir=log_dir, level=log_level)
        logger.info(str(self))

    def __str__(self):
        msg = f"\nSerialTopology Summary\n" +\
            f"---------------------\n"
        msg += f"{'pid':<18s}: {self.pid}\n"
        msg += f"{'log_dir':<18s}: {self.log_dir}\n"
        msg += f"{'logfile':<18s}: {self.logfile}\n"
        return msg


    def _init_log(self, log_dir, level=logging.INFO):
        self.log_dir = "./" if log_dir is None else log_dir
        if not os.path.isdir(self.log_dir):
            os.makedirs(self.log_dir)
        self.logfile = f"{self.log_dir}/log.serial.out"

        logger.setLevel(level=level)
        formatter = SimpleFormatter(fmt="[%(relativeCreated)d s] [%(levelname)-7s] %(message)s")
        handler = logging.FileHandler(self.logfile, mode="w+")
        handler.setLevel(level=level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    @property
    def is_root(self) -> bool:
        return True

    def barrier(self) -> None:
        pass

    def gather(self, local_array):
        return local_array

    # since the communication doesn't need to happen
    # we can assume these results already exist
    # this is just for code compatibility without mpi
    def sum(self, local_array, result_buffer):
        result_buffer[:] = local_array

    def max(self, local_array, result_buffer):
        result_buffer[:] = local_array

    def min(self, local_array, result_buffer):
        result_buffer[:] = local_array

    def any(self, local_array, result_buffer):
        result_buffer[:] = local_array



import os
import yaml
import logging
from typing import Optional
import fsspec

import xarray as xr
import gribapi

logger = logging.getLogger("ufs2arco")

class NOAAGribForecastData:
    """
    Generic access patterns for forecast datasets stored somewhere in Grib

    Classes that inherit this need their own:
        * :attr:`available_level` attribute / property
        * :meth:`_build_path` method
        * :attr:`static_vars`
        * :attr:`sample_dims`
        * :attr:`horizontal_dims`
        * :attr:`file_suffixes`
    """

    # fill these out per subclass (unique to base source)
    file_suffixes = tuple()

    @property
    def _fsname(self) -> str:
        """hrrr, gefs, or gfs"""

        if "gefs" in self.name.lower():
            return "gefs"
        elif "gfs" in self.name.lower():
            return "gfs"
        elif "hrrr" in self.name.lower():
            return "hrrr"
        else:
            raise NotImplementedError(f"ufs2arco.sources.NOAAGribForecastData could not recognize forecast system with name = {self.name}")

    @property
    def available_variables(self) -> tuple:
        return tuple(self._varmeta.keys())

    def __init__(
        self,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
        slices: Optional[dict] = None,
    ) -> None:
        path = os.path.join(
            os.path.dirname(__file__),
            f"reference.{self._fsname}.yaml",
        )
        with open(path, "r") as f:
            self._varmeta = yaml.safe_load(f)
        super().__init__(
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
            slices=slices,
        )

    def _open_static_vars(self, dims) -> bool:
        """Do this once per t0, ensemble member"""
        cond = True
        for key, val in dims.items():
            if key != "t0":
                cond = cond and val == getattr(self, key)[0]
        return cond

    def _open_local(self, dims, file_suffix, cache_dir):

        path = self._build_path(
            **dims,
            file_suffix=file_suffix,
        )
        kw = {"s3": {"anon": True}} if "s3://" in path else {}
        try:
            local_file = fsspec.open_local(
                path,
                filecache={"cache_storage": cache_dir},
                **kw,
            )
        except:
            local_file = None
            logger.warning(
                f"{self.name}: Trouble finding the file: {path}\n\t" +
                f"dims = {dims}, file_suffix = {file_suffix}"
            )
        return local_file

    def open_sample_dataset(
        self,
        dims: dict,
        open_static_vars: bool,
        cache_dir: Optional[str] = None,
    ) -> xr.Dataset:

        # 1. cache the grib files for this date, member, fhr
        cached_files = {}
        for suffix in self.file_suffixes:
            path = self._build_path(
                **dims,
                file_suffix=suffix,
            )
            if cache_dir is not None:
                cached_files[suffix] = self._open_local(dims, suffix, cache_dir)
            else:
                cached_files[suffix] = path

        # 2. read data arrays from those files
        dsdict = {}
        osv = open_static_vars or self._open_static_vars(dims)
        variables = self.variables if osv else self.dynamic_vars
        we_got_the_data = all(val is not None for val in cached_files.values())
        if we_got_the_data:
            for varname in variables:
                dslist = []
                # Note that when there are 2 file suffixes, we always try to read from both
                # because sometimes (e.g. GFS humidity), the variable is in one, sometimes
                # it's in the other. This is more straightforward, and not "too bad".
                for suffix in self._varmeta[varname]["file_suffixes"]:
                    try:
                        thisvar = self._open_single_variable(
                            dims=dims,
                            varname=varname,
                            file=cached_files[suffix],
                        )
                    except:
                        thisvar = None
                    dslist.append(thisvar)
                dslist = [xds for xds in dslist if xds is not None]
                if len(dslist) == 1:
                    dsdict[varname] = dslist[0]
                elif len(dslist) > 1:
                    dsdict[varname] = xr.merge(dslist)
                else:
                    logger.warning(
                        f"{self.name}: Could not find {varname}, will stop reading variables for this sample\n\t" +
                        f"dims = {dims}, file_suffixes = {self._varmeta[varname]['file_suffixes']}"
                    )
                    dsdict = {}
                    break

        # the dataset is either full or completely empty if we had trouble
        xds = xr.Dataset(dsdict)
        if len(xds) > 0:
            xds = self.apply_slices(xds)
        return xds


    def _open_single_variable(
        self,
        dims: dict,
        varname: str,
        file: fsspec.spec.AbstractFileSystem,
    ) -> xr.DataArray:
        """
        Open a single variable from a GRIB file.

        Args:
            file (fsspec.spec.AbstractFileSystem): The file to read.
            varname (str): The variable name to extract.

        Returns:
            xr.DataArray: The extracted variable as an xarray DataArray.
        """
        fbk = self._varmeta[varname]["filter_by_keys"]
        try:
            xds = xr.open_dataset(
                file,
                engine="cfgrib",
                filter_by_keys=fbk,
                decode_timedelta=True,
            )
        except gribapi.errors.WrongLengthError:
            logger.warning(f"{self.name}: gribapi.errors.WrongLengthError for varname = {varname} at dims = {dims}")
            return None

        if "original_name" in self._varmeta[varname]:
            og = self._varmeta[varname]["original_name"]
            xds = xds.rename({og: varname})
            xds[varname].attrs["original_name"] = og
        xda = xds[varname]

        if "isobaricInhPa" in xds.coords:
            if len(xda.dims) < 3:
                vv = xda["isobaricInhPa"].values
                xda = xda.expand_dims({"isobaricInhPa": [vv]})

        if fbk["typeOfLevel"] == "heightAboveGround":
            if "original_name" in self._varmeta[varname]:
                level = fbk["level"]
                xda.attrs["long_name"] = f"{level} metre " + xda.long_name

        elif fbk["typeOfLevel"] == "surface":
            if self._varmeta[varname].get("original_name", "") == "t":
                xda.attrs["long_name"] += " at surface"

            if xda.attrs["GRIB_stepType"] == "accum":
                xda.attrs["long_name"] += " accumulated over forecast"

            elif xda.attrs["GRIB_stepType"] == "avg":
                xda.attrs["long_name"] = "Time-mean " + xda.attrs["long_name"]

        elif fbk["typeOfLevel"] in ("lowCloudLayer", "middleCloudLayer", "highCloudLayer"):
            full = fbk["typeOfLevel"].replace("CloudLayer", "")
            new = f"{full[0]}cc"
            xda.attrs["long_name"] = xda.long_name.replace("Total", full.capitalize())


        for v in [fbk["typeOfLevel"], "number"]:
            if v in xda.coords and v not in xda.dims:
                xda = xda.drop_vars(v)

        xds = xda.to_dataset(name=varname)
        for key, val in self.rename.items():
            if key in xds:
                xds = xds.rename({key: val})

        if varname in self.static_vars:
            for key in ["lead_time", "valid_time"]:
                if key in xds:
                    xds = xds.drop_vars(key)

            xds = xds.expand_dims("t0")
        else:

            # handle vertical levels
            if "level" in xds and self.levels is not None:
                level_selection = [l for l in self.levels if l in xds.level.values]
                if len(level_selection) == 0:
                    # we don't select vertical levels available in this file
                    return None
                xds = xds.sel(level=level_selection, **self._level_sel_kwargs)

            # handle potential ensemble member dimension
            # note that we do this first so that the dims work out in order:
            # t0, fhr, member, level, **horizontal_dims
            if "member" in dims:
                # Note that the description is only known to be true for GEFS...
                # but I'll just leave it for now
                xds = xds.expand_dims("member")
                xds["member"] = xr.DataArray(
                    [dims["member"]],
                    coords={"member": [dims["member"]]},
                    dims=("member",),
                    attrs={
                        "description": "ID=0 comes from gecXX files, ID>0 comes from gepXX files",
                        "long_name": "ensemble member ID",
                    },
                )

            # handle lead_time/fhr coordinates
            xds = xds.expand_dims(["t0", "lead_time"])
            xds["fhr"] = xr.DataArray(
                [int(lt / 1e9 / 3600) for lt in xds["lead_time"].values],
                coords=xds["lead_time"].coords,
                attrs={
                    "long_name": "hours since initial time",
                    "units": "integer hours",
                },
            )
            xds = xds.swap_dims({"lead_time": "fhr"})

            # recreate valid_time, since it's not always there
            valid_time = xds["t0"] + xds["lead_time"]
            if "valid_time" in xds:
                xds["valid_time"] = xds["valid_time"].expand_dims(["t0", "fhr"])
                assert valid_time.squeeze() == xds.valid_time.squeeze()
                xds = xds.drop_vars("valid_time")

            xds["valid_time"] = valid_time
            xds = xds.set_coords("valid_time")

        return xds[varname]


    def open_grib(
        self,
        dims,
        file_suffix,
        cache_dir,
        **kwargs,
    ) -> xr.Dataset:
        """
        Open a single GRIB file.

        Args:
            dims (dict): e.g. {"t0": "2020-01-01T00", "fhr": 0}
            file_suffix (str): e.g. "a", "b", "prs", etc
            cache_dir (str): path to caching the grib file
            kwargs: passed to xarray.open_dataset

        Returns:
            xr.Dataset: with the contents of the grib file, after some minimal postprocessing
        """

        file = self._open_local(dims, file_suffix, cache_dir)
        if file is None:
            return None

        if "decode_timedelta" not in kwargs:
            kwargs["decode_timedelta"] = True
        xds = xr.open_dataset(
            file,
            engine="cfgrib",
            **kwargs,
        )
        if "isobaricInhPa" in xds.coords:
            if len(xds.dims) < 3:
                vv = xds["isobaricInhPa"].values
                xds = xds.expand_dims({"isobaricInhPa": [vv]})

        for key, val in self.rename.items():
            if key in xds:
                xds = xds.rename({key: val})

        # handle potential ensemble member dimension
        # note that we do this first so that the dims work out in order:
        # t0, fhr, member, level, **horizontal_dims
        if "member" in dims:
            # Note that the description is only known to be true for GEFS...
            # but I'll just leave it for now
            xds = xds.expand_dims("member")
            xds["member"] = xr.DataArray(
                [dims["member"]],
                coords={"member": [dims["member"]]},
                dims=("member",),
                attrs={
                    "description": "ID=0 comes from gecXX files, ID>0 comes from gepXX files",
                    "long_name": "ensemble member ID",
                },
            )

        # handle lead_time/fhr coordinates
        xds = xds.expand_dims(["t0", "lead_time"])
        xds["fhr"] = xr.DataArray(
            [int(lt / 1e9 / 3600) for lt in xds["lead_time"].values],
            coords=xds["lead_time"].coords,
            attrs={
                "long_name": "hours since initial time",
                "units": "integer hours",
            },
        )
        xds = xds.swap_dims({"lead_time": "fhr"})

        # recreate valid_time, since it's not always there
        valid_time = xds["t0"] + xds["lead_time"]
        if "valid_time" in xds:
            xds["valid_time"] = xds["valid_time"].expand_dims(["t0", "fhr"])
            assert valid_time.squeeze() == xds.valid_time.squeeze()
            xds = xds.drop_vars("valid_time")

        xds["valid_time"] = valid_time
        xds = xds.set_coords("valid_time")

        return xds
